use anyhow::anyhow;

#[cfg(target_os = "wasi")]
use sonic_channel_wasi::{Dest, IngestChannel, PushRequest, QueryRequest, SearchChannel, SonicChannel};
#[cfg(not(target_os = "wasi"))]
use sonic_channel::{Dest, IngestChannel, PushRequest, QueryRequest, SearchChannel, SonicChannel};
use crate::config::Config;

pub struct IngestSonic{
    ingest_channel: IngestChannel
}

impl IngestSonic{
    pub fn new() -> anyhow::Result<IngestSonic> {
        let ingest_channel = IngestChannel::start(
            format!("{}:{}", Config::global().sonic_search_addr, Config::global().sonic_search_port),
            Config::global().sonic_search_password.clone()
        )?;
        Ok(IngestSonic {
            ingest_channel,
        })
    }
    
    pub fn stop_ingest(&mut self) -> anyhow::Result<()>{
        self.ingest_channel.quit().map_err(|e| anyhow!("Unable to quit the sonic ingest channel"))?;
        Ok(())
    }

    pub fn ingest_document(&self, collection: &str, bucket: &str, object: &str, text: &str) -> anyhow::Result<()> {
        // TODO may be we need to store ingested email names and make a queue  
        let contents = Self::slice_string_into_chunks(&*text);
        for c in contents {
            let dest = Dest::col_buc(collection, bucket).obj(object);
            let p = PushRequest::new(dest, c.to_string());
            self.ingest_channel.push(p)?;        
        }
        
        Ok(())
    }

    fn slice_string_into_chunks(s: &str) -> Vec<&str> {
        let mut chunks = Vec::new();
        let mut start = 0;
        let len = s.len();

        while start < len {
            let end = if start + 15000 > len {
                len
            } else {
                let mut end = start + 15000;
                while !s.is_char_boundary(end) && end > start {
                    end -= 1;
                }
                end
            };

            if start != end {
                chunks.push(&s[start..end]);
            }
            start = end;
        }

        chunks
    }
}

pub struct SearchSonic{}

impl SearchSonic{
    pub fn search_document(collection: &str, bucket: &str, query: &str, limit: i32, offset: i32) -> anyhow::Result<Vec<String>> {
        let search = SearchChannel::start(
            format!("{}:{}", Config::global().sonic_search_addr, Config::global().sonic_search_port), 
            Config::global().sonic_search_password.clone()
        )?;
        let results: Vec<String> = search.query(QueryRequest::new(
            Dest::col_buc(collection, bucket),
            query
        ).limit(limit as usize).offset(offset as usize)).unwrap_or_else(|_| { vec![] });
        search.quit()?;
        Ok(results)
    }
}